package ru.ulstu.extractor.assessment.service;

import org.springframework.stereotype.Service;
import ru.ulstu.extractor.assessment.model.Assessment;
import ru.ulstu.extractor.rule.model.AssessmentException;
import ru.ulstu.extractor.rule.model.DbRule;
import ru.ulstu.extractor.rule.service.AntecedentValueService;
import ru.ulstu.extractor.rule.service.DbRuleService;
import ru.ulstu.extractor.rule.service.FuzzyInferenceService;
import ru.ulstu.extractor.ts.model.TimeSeries;
import ru.ulstu.extractor.ts.service.TimeSeriesService;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Service
public class AssessmentService {
    private final DbRuleService ruleService;
    private final AntecedentValueService antecedentValueService;
    private final TimeSeriesService timeSeriesService;
    private final FuzzyInferenceService fuzzyInferenceService;

    public AssessmentService(DbRuleService ruleService,
                             AntecedentValueService antecedentValueService,
                             TimeSeriesService timeSeriesService,
                             FuzzyInferenceService fuzzyInferenceService) {
        this.ruleService = ruleService;
        this.antecedentValueService = antecedentValueService;
        this.timeSeriesService = timeSeriesService;
        this.fuzzyInferenceService = fuzzyInferenceService;
    }

    public List<Assessment> getAssessments(Integer branchId) {
        List<TimeSeries> timeSeries = timeSeriesService.getByBranch(branchId);
        List<DbRule> dbRules = ruleService.getList();
        try {
            return getAssessments(dbRules, timeSeries);
        } catch (AssessmentException ex) {
            ex.printStackTrace();
            return new ArrayList<>();
        }
    }

    public List<Assessment> getSingleAssessment(Integer branchId) throws AssessmentException {
        List<TimeSeries> timeSeries = timeSeriesService.getByBranch(branchId);
        List<DbRule> dbRules = ruleService.getList();
        return getSingleAssessment(dbRules, timeSeries);
    }

    private List<Assessment> getSingleAssessment(List<DbRule> dbRules, List<TimeSeries> timeSeries) throws AssessmentException {
        Map<String, Double> variableValues = new HashMap<>();
        timeSeries.forEach(ts -> variableValues.put(ts.getTimeSeriesType().name(), timeSeriesService.getLastTimeSeriesTendency(ts)));
        return fuzzyInferenceService.getFuzzyInference(dbRules,
                antecedentValueService.getList(),
                variableValues);
    }

    private List<Assessment> getAssessments(List<DbRule> dbRules, List<TimeSeries> timeSeries) {
        return dbRules
                .stream()
                .flatMap(dbRule -> {
                    Map<String, Double> variableValues = new HashMap<>();
                    timeSeries
                            .stream()
                            .filter(ts -> ts.getTimeSeriesType() == dbRule.getFirstAntecedent()
                                    || ts.getTimeSeriesType() == dbRule.getSecondAntecedent())
                            .forEach(ts -> variableValues.put(ts.getTimeSeriesType().name(), timeSeriesService
                                    .getLastTimeSeriesTendency(ts)));
                    return fuzzyInferenceService.getFuzzyInference(List.of(dbRule),
                            antecedentValueService.getList(),
                            variableValues).stream();
                })
                .sorted(Comparator.comparing(Assessment::getDegree))
                .collect(Collectors.toList());
    }
}
