/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.branch.service;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.ulstu.extractor.branch.model.Branch;
import ru.ulstu.extractor.branch.model.IndexingStatus;
import ru.ulstu.extractor.branch.repository.BranchRepository;
import ru.ulstu.extractor.commit.model.Commit;
import ru.ulstu.extractor.commit.service.CommitService;
import ru.ulstu.extractor.core.BaseEntity;
import ru.ulstu.extractor.gitrepository.model.GitRepository;
import ru.ulstu.extractor.ts.service.TimeSeriesService;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
public class BranchService {
    private final static Logger LOG = LoggerFactory.getLogger(BranchService.class);
    private final BranchRepository branchRepository;
    private final CommitService commitService;
    private final TimeSeriesService timeSeriesService;

    public BranchService(BranchRepository branchRepository,
                         CommitService commitService,
                         TimeSeriesService timeSeriesService) {
        this.branchRepository = branchRepository;
        this.commitService = commitService;
        this.timeSeriesService = timeSeriesService;
    }

    @Transactional
    public Branch save(Branch branch) {
        return branchRepository.save(branch);
    }

    @Transactional
    public Branch save(Branch branch, List<Commit> commits) {
        LOG.debug("Start save {} branch with {} commits ", branch.getName(), commits.size());
        LOG.debug("Current branch contains {} commits ", branch.getCommits().size());
        List<Integer> commitsToRemoveIds = branch.getCommits().stream().map(BaseEntity::getId).collect(Collectors.toList());
        branch.setCommits(commitService.save(commits));
        LOG.debug("Save branch {} ", branch.getName());
        branch = branchRepository.save(branch);
        LOG.debug("Clear {} commits ", commitsToRemoveIds.size());
        commitService.deleteWithEmptyIds();
        return branch;
    }

    @Transactional
    public Branch addCommits(Branch branch, List<Commit> commits) {
        LOG.debug("Start add commits to {} branch with {} commits ", branch.getName(), commits.size());
        branch.getCommits().addAll(commitService.save(commits));
        LOG.debug("Save branch {} ", branch.getName());
        return save(branch);
    }

    public Branch findByRepositoryAndName(GitRepository gitRepository, String branchName) {
        return branchRepository.findByGitRepositoryAndName(gitRepository, branchName);
    }

    public Branch findByRepositoryAndNameOrCreate(GitRepository gitRepository, String branchName) {
        Branch branch = branchRepository.findByGitRepositoryAndName(gitRepository, branchName);
        return branch == null
                ? branchRepository.save(new Branch(gitRepository, branchName))
                : branch;
    }

    public Optional<Branch> findByBranchId(Integer branchId) {
        return branchRepository.findById(branchId);
    }

    public List<Branch> findAll() {
        return branchRepository.findAll();
    }

    public List<Branch> findAllByIndexingStatus(IndexingStatus indexingStatus) {
        return branchRepository.findAllByIndexingStatus(indexingStatus);
    }

    @Transactional
    public Branch updateStatus(Branch branch, IndexingStatus indexingStatus) {
        branch.setIndexingStatus(indexingStatus);
        return branchRepository.save(branch);
    }

    public List<Branch> findAllValid() {
        return findAll()
                .stream()
                .filter(timeSeriesService::isBranchContainsAllTimeSeries)
                .collect(Collectors.toList());
    }
}
