package ru.ulstu.extractor.ts.service;

import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.branch.model.Branch;
import ru.ulstu.extractor.http.HttpService;
import ru.ulstu.extractor.http.JsonTimeSeries;
import ru.ulstu.extractor.http.SmoothingTimeSeries;
import ru.ulstu.extractor.ts.model.TimeSeries;
import ru.ulstu.extractor.ts.model.TimeSeriesType;
import ru.ulstu.extractor.ts.model.TimeSeriesValue;
import ru.ulstu.extractor.ts.repository.TimeSeriesRepository;
import ru.ulstu.extractor.ts.repository.TimeSeriesValueRepository;
import ru.ulstu.extractor.ts.util.TimeSeriesDateMapper;
import ru.ulstu.extractor.util.JsonUtils;

import javax.transaction.Transactional;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
public class TimeSeriesService {
    private final static Logger LOG = LoggerFactory.getLogger(TimeSeriesService.class);
    private final static int MIN_TIME_SERIES_LENGTH = 5;
    private final static double DEFAULT_TIME_SERIES_TENDENCY = 0.0;
    private final TimeSeriesRepository timeSeriesRepository;
    private final TimeSeriesValueRepository timeSeriesValueRepository;
    private final TimeSeriesDateMapper.TimeSeriesInterval timeSeriesInterval = TimeSeriesDateMapper.TimeSeriesInterval.HOUR;
    private final HttpService httpService;
    private final static String TIME_SERIES_SAVE_SERVICE_URL = "http://time-series.athene.tech/api/1.0/add-time-series?setKey=git-extractor";
    private final static String TIME_SERIES_TENDENCY_URL = "http://time-series.athene.tech/api/1.0/getSpecificMethodSmoothed";
    private final static String TIME_SERIES_GROUPED_TENDENCIES_URL = "http://time-series.athene.tech/api/1.0/getGroupedTendencies";

    public TimeSeriesService(TimeSeriesRepository timeSeriesRepository,
                             TimeSeriesValueRepository timeSeriesValueRepository,
                             HttpService httpService) {
        this.timeSeriesRepository = timeSeriesRepository;
        this.timeSeriesValueRepository = timeSeriesValueRepository;
        this.httpService = httpService;
    }

    @Transactional
    public TimeSeries save(String timeSeriesName, Branch branch, TimeSeriesType timeSeriesType, List<TimeSeriesValue> timeSeriesValues) {
        LOG.debug("Начало сохранения временного ряда {} с {} значениями", timeSeriesName, timeSeriesValues.size());
        final TimeSeries timeSeries = findOrCreate(timeSeriesName, branch, timeSeriesType);
        List<TimeSeriesValue> timeSeriesValuesToRemove = timeSeries.getValues();
        timeSeries.setValues(timeSeriesValues);
        LOG.debug("Сохранение временного ряда {} ", timeSeries.getName());
        TimeSeries savedTimeSeries = timeSeriesRepository.save(timeSeries);
        LOG.debug("Очистить {} точек временного ряда", timeSeriesValuesToRemove.size());
        timeSeriesValueRepository.deleteAll(timeSeriesValuesToRemove);
        sendToTimeSeriesService(savedTimeSeries);
        return savedTimeSeries;
    }

    private TimeSeries findOrCreate(String timeSeriesName, Branch branch, TimeSeriesType timeSeriesType) {
        Optional<TimeSeries> maybeTimeSeries = timeSeriesRepository.findByName(timeSeriesName);
        if (maybeTimeSeries.isPresent()) {
            LOG.debug("TimeSeries {} exists.", maybeTimeSeries.get().getName());
            return maybeTimeSeries.get();
        }
        return timeSeriesRepository.save(new TimeSeries(timeSeriesName, branch, timeSeriesType));
    }

    public TimeSeries findByBranchAndTypeOrCreate(Branch branch, TimeSeriesType timeSeriesType) {
        Optional<TimeSeries> maybeTimeSeries = timeSeriesRepository.getTimeSeriesByBranchAndTimeSeriesType(branch, timeSeriesType);
        if (maybeTimeSeries.isPresent()) {
            LOG.debug("TimeSeries {} exists.", maybeTimeSeries.get().getName());
            return maybeTimeSeries.get();
        }
        return timeSeriesRepository.save(new TimeSeries(branch, timeSeriesType));
    }

    /**
     * Сохранить список временных рядов
     *
     * @param timeSeries список ВР
     * @return сохраненный список ВР
     */
    @Transactional
    public List<TimeSeries> save(List<TimeSeries> timeSeries, Branch branch) {
        List<TimeSeries> results = new ArrayList<>();
        for (TimeSeries ts : timeSeries) {
            results.add(save(ts.getName(), branch, ts.getTimeSeriesType(), ts.getValues()));
        }
        return results;
    }

    public TimeSeriesDateMapper.TimeSeriesInterval getTimeSeriesInterval() {
        return timeSeriesInterval;
    }

    private void sendToTimeSeriesService(TimeSeries timeSeries) {
        new Thread(() -> {
            try {
                httpService.post(TIME_SERIES_SAVE_SERVICE_URL, new JSONObject(new JsonTimeSeries(timeSeries)));
                LOG.debug("{} успешно отправлен на сервис временных рядов", timeSeries.getName());
            } catch (Exception ex) {
                LOG.debug(ex.getMessage());
            }
        }).start();
    }

    public List<TimeSeries> getAllTimeSeries() {
        return timeSeriesRepository.findAll();
    }

    public TimeSeries getById(Integer tsId) {
        return timeSeriesRepository.findById(tsId)
                .orElseThrow(() -> new RuntimeException("Временной ряд не найден по идентификатору " + tsId));
    }

    public List<TimeSeriesType> getAllTimeSeriesTypes() {
        return Arrays.asList(TimeSeriesType.values());
    }

    public List<TimeSeries> getByBranch(Integer branchId) {
        return timeSeriesRepository.getTimeSeriesByBranchId(branchId);
    }

    public double getLastTimeSeriesTendency(TimeSeries ts) {
        if (ts != null && ts.getValues().size() > MIN_TIME_SERIES_LENGTH) {
            JSONObject response = httpService.post(TIME_SERIES_TENDENCY_URL, new JSONObject(new SmoothingTimeSeries(normalizeTimeSeries(ts))));
            LOG.debug("Успешно отправлен на сервис сглаживания");
            if (response.has("response") && response.getString("response").equals("empty")) {
                return DEFAULT_TIME_SERIES_TENDENCY;
            }
            JSONArray jsonArray = response.getJSONObject("timeSeries").getJSONArray("values");
            return jsonArray.getJSONObject(jsonArray.length() - 1).getDouble("value") -
                    jsonArray.getJSONObject(jsonArray.length() - 2).getDouble("value");
        }
        return DEFAULT_TIME_SERIES_TENDENCY;
    }

    private TimeSeries normalizeTimeSeries(TimeSeries ts) {
        double sum = ts.getValues().stream().mapToDouble(TimeSeriesValue::getValue).sum();

        if (sum > 0.0d) {
            for (int i = 0; i < ts.getValues().size(); i++) {
                ts.getValues().get(i).setValue(ts.getValues().get(i).getValue() / sum);
            }
        }
        return ts;
    }

    public boolean isBranchContainsAllTimeSeries(Branch b) {
        List<TimeSeries> timeSeries = getByBranch(b.getId());
        return Stream.of(TimeSeriesType.values()).allMatch(type -> timeSeries
                .stream()
                .anyMatch(ts -> type == ts.getTimeSeriesType()));
    }

    public List<TimeSeries> getGroupedTendencies(List<TimeSeries> tsList) {
        JSONArray response = httpService.post(TIME_SERIES_GROUPED_TENDENCIES_URL, new JSONArray(tsList.stream().map(JsonTimeSeries::new).collect(Collectors.toList())));
        LOG.debug("Send to group time series tendencies");
        return JsonUtils.getListOfObjects(response)
                .stream()
                .map(TimeSeries::new)
                .collect(Collectors.toList());
    }
}
